"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var _task, _context, _nextRunTimeEvaluator, _nextRunData, _taskRunningPromise, _firstAttempt;
Object.defineProperty(exports, "__esModule", { value: true });
exports.buildEvaluator = exports.SingleInstanceTaskScheduler = void 0;
const assert_1 = require("assert");
/**
 * A single instance task scheduler with flexible next run time.
 *
 * Stability: 1 - Experimental.
 */
// eslint-disable-next-line @typescript-eslint/no-invalid-void-type
class SingleInstanceTaskScheduler {
    constructor(task, initialContext, options) {
        var _a;
        _task.set(this, void 0);
        _context.set(this, void 0);
        _nextRunTimeEvaluator.set(this, void 0);
        /**
         * `null` indicates there is no next run, `undefined` indicate next run information
         * is going to be set after the current running task end.
         */
        _nextRunData.set(this, null);
        _taskRunningPromise.set(this, null);
        _firstAttempt.set(this, null);
        __classPrivateFieldSet(this, _task, task);
        __classPrivateFieldSet(this, _context, initialContext);
        const nextRunTime = (_a = options === null || options === void 0 ? void 0 : options.nextRunTime) !== null && _a !== void 0 ? _a : null;
        if (typeof nextRunTime === 'function' || nextRunTime === null) {
            __classPrivateFieldSet(this, _nextRunTimeEvaluator, nextRunTime);
        }
        else {
            __classPrivateFieldSet(this, _nextRunTimeEvaluator, buildEvaluator(nextRunTime));
        }
    }
    /**
     * Whether a next run is scheduled.
     */
    get scheduled() {
        return (__classPrivateFieldGet(this, _nextRunData) !== null);
    }
    // Maybe we can consider to return `Date | null | Promise<Date | null>`?
    /**
     * The next run time of the task. `null` indicates that there will be no next run.
     * `undefined` indicates that the next run time is going to be known after the current
     * task end.
     */
    get nextRunTime() {
        let result;
        if (__classPrivateFieldGet(this, _nextRunData) == null) {
            result = __classPrivateFieldGet(this, _nextRunData);
        }
        else {
            result = __classPrivateFieldGet(this, _nextRunData).startTime;
        }
        return result;
    }
    /**
     * Whether the task is currently running.
     */
    get running() {
        return __classPrivateFieldGet(this, _taskRunningPromise) !== null;
    }
    /**
     * Schedule the task to run after a given milliseconds or absolute date time. If
     * the task is already scheduled, it will be re-scheduled.
     * @param startDelayOrTime Start time of the next run. A delay in milliseconds, or an absolute
     *   date time.
     */
    schedule(startDelayOrTime) {
        var _a;
        const prevAttemptNumber = (_a = __classPrivateFieldGet(this, _nextRunData)) === null || _a === void 0 ? void 0 : _a.attemptNumber;
        let startTime;
        let delay;
        if (startDelayOrTime instanceof Date) {
            startTime = startDelayOrTime;
            delay = startDelayOrTime.getTime() - Date.now();
        }
        else {
            delay = startDelayOrTime;
            startTime = new Date(Date.now() + startDelayOrTime);
        }
        this.cancelNextRun();
        __classPrivateFieldSet(this, _nextRunData, {
            startTime,
            timer: setTimeout(this.run.bind(this), delay),
            attemptNumber: prevAttemptNumber !== null && prevAttemptNumber !== void 0 ? prevAttemptNumber : 1
        });
    }
    /**
     * Cancel the next scheduled run. Running task will not be affected.
     */
    cancelNextRun() {
        if (__classPrivateFieldGet(this, _nextRunData) != null) {
            clearTimeout(__classPrivateFieldGet(this, _nextRunData).timer);
        }
        __classPrivateFieldSet(this, _nextRunData, null);
    }
    scheduleWithResult(taskResult, startTime, endTime) {
        var _a, _b, _c, _d, _e, _f;
        if (__classPrivateFieldGet(this, _nextRunTimeEvaluator) == null) {
            __classPrivateFieldSet(this, _nextRunData, null);
        }
        else {
            const thisRunAttemptNumber = (_b = (_a = __classPrivateFieldGet(this, _nextRunData)) === null || _a === void 0 ? void 0 : _a.attemptNumber) !== null && _b !== void 0 ? _b : 1;
            const nextRunRequest = __classPrivateFieldGet(this, _nextRunTimeEvaluator).call(this, taskResult, {
                firstAttemptStartTime: (_d = (_c = __classPrivateFieldGet(this, _firstAttempt)) === null || _c === void 0 ? void 0 : _c.startTime) !== null && _d !== void 0 ? _d : startTime,
                firstAttemptEndTime: (_f = (_e = __classPrivateFieldGet(this, _firstAttempt)) === null || _e === void 0 ? void 0 : _e.endTime) !== null && _f !== void 0 ? _f : endTime,
                attemptNumber: thisRunAttemptNumber,
                startTime,
                endTime
            }, __classPrivateFieldGet(this, _context));
            if (nextRunRequest === null) {
                __classPrivateFieldSet(this, _nextRunData, null);
            }
            else if (__classPrivateFieldGet(this, _nextRunData) !== null) {
                // Next run was not cancelled, so we schedule the next run
                this.schedule(nextRunRequest.startDelayOrTime);
                if (__classPrivateFieldGet(this, _nextRunData) === undefined) {
                    assert_1.strict.fail('nextRunData should not be undefined');
                }
                // Update first attempt metadata and next run attempt number
                if (nextRunRequest.isRetry) {
                    if (thisRunAttemptNumber === 1) {
                        __classPrivateFieldSet(this, _firstAttempt, {
                            startTime,
                            endTime
                        });
                    }
                    __classPrivateFieldGet(this, _nextRunData).attemptNumber = thisRunAttemptNumber + 1;
                }
                else {
                    __classPrivateFieldSet(this, _firstAttempt, null);
                    __classPrivateFieldGet(this, _nextRunData).attemptNumber = 1;
                }
            }
        }
    }
    /**
     * Run the scheduled task immediately without waiting for the task returned value. Next
     * run will be scheduled if configured.
     */
    run() {
        if (__classPrivateFieldGet(this, _nextRunData) === null) {
            __classPrivateFieldSet(this, _nextRunData, undefined);
        }
        if (__classPrivateFieldGet(this, _taskRunningPromise) !== null) {
            return;
        }
        // In case of implementation error, we will just let it throw so that we can notice such error
        // eslint-disable-next-line @typescript-eslint/no-floating-promises
        __classPrivateFieldSet(this, _taskRunningPromise, (async () => {
            try {
                let taskResult;
                const startTime = new Date();
                try {
                    taskResult = {
                        type: 'SUCCESS',
                        returnValue: await __classPrivateFieldGet(this, _task).call(this, __classPrivateFieldGet(this, _context))
                    };
                }
                catch (error) {
                    taskResult = {
                        type: 'ERROR',
                        caughtValue: error
                    };
                }
                const endTime = new Date();
                this.scheduleWithResult(taskResult, startTime, endTime);
                if (taskResult.type === 'ERROR') {
                    throw taskResult.caughtValue;
                }
                return taskResult.returnValue;
            }
            finally {
                __classPrivateFieldSet(this, _taskRunningPromise, null);
            }
        })());
    }
    /**
     * Run the scheduled task immediately and return a promise which will be fulfilled
     * with the task returned value. Next run will be scheduled if configured.
     */
    async runReturnResult() {
        this.run();
        if (__classPrivateFieldGet(this, _taskRunningPromise) === null) {
            assert_1.strict.fail('taskRunningPromise should not be null');
        }
        return await __classPrivateFieldGet(this, _taskRunningPromise);
    }
}
exports.SingleInstanceTaskScheduler = SingleInstanceTaskScheduler;
_task = new WeakMap(), _context = new WeakMap(), _nextRunTimeEvaluator = new WeakMap(), _nextRunData = new WeakMap(), _taskRunningPromise = new WeakMap(), _firstAttempt = new WeakMap();
function buildEvaluator(options) {
    return (result, meta) => {
        let request;
        if (result.type === 'SUCCESS') {
            const onSuccess = options.onSuccess;
            if (onSuccess.type === 'ONE_TIME') {
                request = null;
            }
            else if (onSuccess.type === 'RUN_START_TIME') {
                request = {
                    // TODO Change to return absolute Date object for higher accuracy
                    startDelayOrTime: (meta.startTime.getTime() + onSuccess.delay) - Date.now(),
                    isRetry: false
                };
            }
            else if (onSuccess.type === 'RUN_END_TIME') {
                request = {
                    startDelayOrTime: onSuccess.delay,
                    isRetry: false
                };
            }
            else {
                assert_1.strict.fail('Not implemented onSuccess.type');
            }
        }
        else if (result.type === 'ERROR') {
            if (options.onError === undefined ||
                (options.onError.attempt !== undefined && meta.attemptNumber >= options.onError.attempt)) {
                request = null;
            }
            else {
                request = {
                    startDelayOrTime: options.onError.delay,
                    isRetry: true
                };
            }
        }
        else {
            assert_1.strict.fail('Not implemented onError.type');
        }
        return request;
    };
}
exports.buildEvaluator = buildEvaluator;
